#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char byte;

const char basic_part[] = {
    0x0E, 0x7F, 0x6A,                                   // -> List1
    0x0D, 0xED,                                         // Prog "C8INIT"
    '"', 'C', '8', 'I', 'N', 'I', 'T', '"',
    0x0D, 0xF7, 0x08, 'K', 0x11, '4', '7',              // While K!=47
    0x0D, 0x7F, 0x8F, 0x0E, 'K',                        // GetKey->K
    0x0D, 0xED,                                         // Prog "C8UPDATE"
    '"', 'C', '8', 'U', 'P', 'D', 'A', 'T', 'E', '"',
    0x0D, 0xF7, 0x09,                                   // WhileEnd
    0x00
};

char* longToString(long long value);
char* concat(char* string1, char* string2);
long long pow(int x, int n);

int main(int argc, char** argv){

    FILE *ch8_file, *g1m_file;
    byte *game;
    int ch8_fileSize, g1m_fileSize;
    long long list_entry;
    char *output, *header;

    if(argc != 3){
        printf("CH8TOTXT [Ch8 File] [Output file]");
        return 0;
    }

    // Open the Chip 8 game
    ch8_file = fopen(argv[1], "rb");
    if(!ch8_file){
        printf("Unable to load %s.", argv[1]);
        return -1;
    }

    // Check if the file isn't too big
    fseek(ch8_file, 0L, SEEK_END);
    ch8_fileSize = ftell(ch8_file);
    rewind(ch8_file);

    if(ch8_fileSize > 0xDFF){

        printf("The following file is too big : %s.", argv[1]);
        fclose(ch8_file);
        return -1;
    }

    // Load the game into memory
    game = malloc(ch8_fileSize * sizeof(byte));
    if(game == NULL){
        printf("Memory Error");
        fclose(ch8_file);
        return -1;
    }

    if(fread(game, sizeof(byte), ch8_fileSize, ch8_file) != ch8_fileSize){
        printf("The following file is unreadable : %s.", argv[1]);
        fclose(ch8_file);
        return -1;
    }

    // We don't need to let the Chip 8 file open anymore
    fclose(ch8_file);

    // Open the text file
    g1m_file = fopen(argv[2], "w+");
    if(!g1m_file){
        printf("Unable to write %s.", argv[2]);
        return -1;
    }

    // Transform the game into a basic program
    list_entry = 0;

    output = "{";
    for(int i = 0; i < ch8_fileSize; i++){

        list_entry <<= 8;
        list_entry |= game[i];

        if(i % 4 == 3 || i == (ch8_fileSize - 1)){

            while(i >= (ch8_fileSize - 1) && (i % 4 != 3)){
                list_entry <<= 8;
                list_entry |= 0;
                i++;
            }

            output = concat(output, longToString(list_entry));

            if(i < (ch8_fileSize - 1))
                output = concat(output, ",");

            list_entry = 0;
        }
    }

    // Add all the other basic instructions
    output = concat(output, (char*)basic_part);

    // Free the binary file
    free(game);

    // Calculate the g1m file's size (needed for the header)
    g1m_fileSize = strlen(output) + 0x56 + (4 - strlen(output) % 4) + 2;

    // Make the header (configured for one basic program only)
    header = malloc(0x56 * sizeof(char));

    strcpy(header, "USBPower");                     // File Identifier
    header[8] = 0x31;                               // File type identifier
    for(int i = 0; i < 5; i++){                     // {0x00, 0x10, 0x00, 0x10, 0x00}
        header[9 + i] = 0x10 * (i % 2);
    }
    header[0x0F] = 0x01;                            // 0x01
    header[0x13] = (g1m_fileSize      ) & 0x0FF;    // The file's size (in big endian format)
    header[0x12] = (g1m_fileSize >>  8) & 0x0FF;
    header[0x11] = (g1m_fileSize >> 16) & 0x0FF;
    header[0x10] = (g1m_fileSize >> 24) & 0x0FF;

    header[0x0E] = (g1m_fileSize + 0x41) & 0x0FF;   // Control byte
    header[0x14] = (g1m_fileSize + 0xB8) & 0x0FF;
    /* The bytes between 0x15 and 0x1E are insignificant */
    header[0x1E] = 0x00;                            // The number of objects
    header[0x1F] = 0x01;

    // All the bytes in the main header are inversed
    for(int i = 0; i < 32; i++)
        header[i] = 0xFF - header[i];

    // Make the sub header
    strcpy(header + 0x20, "PROGRAM");               // ID of the part
    for(int i = 0; i < 9; i++)
        header[i + 0x27] = 0;
    header[0x30] = 0x00;                            // Number of sub parts
    header[0x31] = 0x00;
    header[0x32] = 0x00;
    header[0x33] = 0x01;

    // Make the subpart header
    strcpy(header + 0x34, "system");                // Directory name (system for basic programs)
    header[0x3A] = 0x00;
    header[0x3B] = 0x00;
    strcpy(header + 0x3C, "C8");                    // Name of the program (start with C8)
    for(int i = 0; i < 6; i++){
        if(argv[1][i] != '.')
            header[0x3E + i] = argv[1][i];
        else
            header[0x3E + i] = ' ';
    }
    header[0x44] = 0x01;                            // Directory type
    for(int i = 0; i < 4; i++)                      // Subpart's size (start just after the padding, so we have to count the password + the bytes at the end)
        header[0x45 + i] = ((strlen(output) + 12 + (4 - strlen(output) % 4)) >> (8 * (3 - i))) & 0x0FF;
    for(int i = 0; i < 3; i++)                      // Padding
        header[0x49 + i] = 0;
    for(int i = 0; i < 10; i++)                     // Password
        header[0x4C + i] = 0;

    // Write the text file
    fwrite(header, sizeof(char), 0x56, g1m_file);
    fputs(output, g1m_file);
    for(int i = 0; i < (4 - strlen(output) % 4) + 2; i++)
        fputc(0, g1m_file);

    free(output);
    fclose(g1m_file);

    return 0;
}

char* longToString(long long value){
    int length = 1;
    long long tmp = value;
    char* output;

    // Calculate the length of the string
    while(tmp >= 10){
        length ++;
        tmp /= 10;
    };

    // Create the string
    output = malloc((length + 1) * sizeof(char));
    output[length] = '\0';

    // Add all the character
    for(int i = 0; i < length; i++){

        tmp = pow(10, length - i - 1);
        output[i]  = (long long)(value / tmp) % 10;
        output[i] += '0';                           // Convert it to text
    }

    return output;
}
long long pow(int x, int n){
    long long output = 1;

    for(int i = 0; i < n; i++)
        output *= x;

    return output;
}
char* concat(char* string1, char* string2){

    if(string1 != NULL && string2 != NULL){
        char *result = malloc(strlen(string1) + strlen(string2) + 1);
        strcpy(result, string1);
        strcat(result, string2);

        free(string1);
        free(string2);

        return result;
    }

    return NULL;
}

